"""Tools for fitting General Linear Models (GLMs) and performing permutation tests."""

from osl_dynamics.glm.base import DesignConfig, GLM
from osl_dynamics.glm.permutation import MaxStatPermutation
